@@ -56,6 +56,8 @@ FETCH_MODE_ASYNC = b"async"
 
 ENCRYPTION_ALGORITHM= "AES256"
 
+DEFAULT_BOS_DOMAIN_SUFFIX = b'bcebos.com'
+
 
 class UploadTaskHandle:
     """
@@ -450,50 +452,65 @@ class BosClient(BceBaseClient):
         :return:
             **HttpResponse Class**
         """
+        params={b'replication': b''}
+        if "id" in replication:
+            params[b"id"] = compat.convert_to_bytes(replication["id"])
         return self._send_request(http_methods.PUT,
                            bucket_name,
                            body=json.dumps(replication,
                                            default=BosClient._dump_acl_object),
                            headers={http_headers.CONTENT_TYPE: http_content_types.JSON},
-                           params={b'replication': b''},
+                           params=params,
                            config=config)
 
     @required(bucket_name=(bytes, str))
-    def get_bucket_replication(self, bucket_name, config=None):
+    def get_bucket_replication(self, bucket_name, id=None, config=None):
         """
         Get configuration of cross-region replication 
 
         :type bucket: string
         :param bucket: None
 
+        :type id: string
+        :param id: replication rule id
+
         :return:
             **HttpResponse Class**
         """
+        params={b'replication': b''}
+        if id is not None:
+            params[b"id"] = compat.convert_to_bytes(id)
         return self._send_request(http_methods.GET,
                            bucket_name,
-                           params={b'replication': b''},
+                           params=params,
                            config=config)
 
 
     @required(bucket_name=(bytes, str))
-    def delete_bucket_replication(self, bucket_name, config=None):
+    def delete_bucket_replication(self, bucket_name, id=None, config=None):
         """
         Delete configuration of cross-region replication and close it 
 
         :type bucket: string
         :param bucket: None
 
+        :type id: string
+        :param id: replication rule id
+
         :return:
             **HttpResponse Class**
         """
+        params={b'replication': b''}
+        if id is not None:
+            params[b"id"] = compat.convert_to_bytes(id)
         return self._send_request(http_methods.DELETE,
                            bucket_name,
-                           params={b'replication': b''},
+                           params=params,
                            config=config)
 
 
     @required(bucket_name=(bytes, str))
-    def get_bucket_replication_progress(self, bucket_name, config=None):
+    def get_bucket_replication_progress(self, bucket_name, id=None, config=None):
         """
         Get status of cross-region replication,for exapmle 'historyReplicationPercent',
         'latestReplicationTime'
@@ -501,12 +518,34 @@ class BosClient(BceBaseClient):
         :type bucket: string
         :param bucket: None
 
+        :type id: string
+        :param id: replication rule id
+
+        :return:
+            **HttpResponse Class**
+        """
+        params={b'replicationProgress': b''}
+        if id is not None:
+            params[b"id"] = compat.convert_to_bytes(id)
+        return self._send_request(http_methods.GET,
+                           bucket_name,
+                           params=params,
+                           config=config)
+    
+    @required(bucket_name=(bytes, str))
+    def list_bucket_replication(self, bucket_name, config=None):
+        """
+        list configuration of cross-region replication rule
+
+        :type bucket: string
+        :param bucket: None
+
         :return:
             **HttpResponse Class**
         """
         return self._send_request(http_methods.GET,
                            bucket_name,
-                           params={b'replicationProgress': b''},
+                           params={b'replication': b'', b'list': b''},
                            config=config)
 
     @required(bucket_name=(bytes, str), inventory=(dict))
@@ -662,11 +701,13 @@ class BosClient(BceBaseClient):
         :type options: dict
         :param options: None
 
+        :param is_official_domain: default use not official domain,example: bucket.bj.bcebos.com
+
         :return:
             **URL string**
         """
         key = compat.convert_to_bytes(key)
-        config = self._merge_config(config)
+        config = self._merge_config(config, bucket_name)
         headers = headers or {}
         params = params or {}
 
@@ -684,6 +725,11 @@ class BosClient(BceBaseClient):
         path = self._get_path(config, bucket_name, key)
         if httpmethod != http_methods.GET and httpmethod != http_methods.HEAD:
             headers_to_sign = set([b'host'])
+
+        # Compatible with STS request acquisition
+        if config.security_token is not None:
+            params[http_headers.STS_SECURITY_TOKEN.lower()] = config.security_token
+
         params[http_headers.AUTHORIZATION.lower()] = bce_v1_signer.sign(
             config.credentials,
             httpmethod,
@@ -693,6 +739,7 @@ class BosClient(BceBaseClient):
             timestamp,
             expiration_in_seconds,
             headers_to_sign)
+        
         return b"%s://%s%s?%s" % (compat.convert_to_bytes(endpoint_protocol.name),
                                  full_host,
                                  path,
@@ -906,7 +953,7 @@ class BosClient(BceBaseClient):
         return True
 
     @required(bucket_name=(bytes, str), key=(bytes, str))
-    def get_object(self, bucket_name, key, range=None, config=None):
+    def get_object(self, bucket_name, key, range=None, traffic_limit=None, config=None):
         """
 
         :param bucket_name:
@@ -916,11 +963,18 @@ class BosClient(BceBaseClient):
         :return:
         """
         key = compat.convert_to_bytes(key)
+        if len(key) == 0 or key.startswith(b"/"):
+            raise BceClientError("Key can not be empty or start with '/' .")
+        range_header = BosClient._get_range_header_dict(range)
+        if traffic_limit is not None:
+            if range_header is None:
+                range_header = {}
+            range_header[http_headers.BOS_TRAFFIC_LIMIT] = traffic_limit
         return self._send_request(
             http_methods.GET,
             bucket_name,
             key,
-            headers=BosClient._get_range_header_dict(range),
+            headers=range_header,
             config=config,
             body_parser=BosClient._parse_bos_object)
 # restore object
@@ -937,9 +991,10 @@ class BosClient(BceBaseClient):
         headers = {}
         if days is not None:
             headers[http_headers.BOS_RESTORE_DAYS] = days
-        if tier not in ("Standard", "Expedited"):
-            raise ValueError('valid tier:%s for restore_object.The valid value is \"Standard\" and \"Expedited\"' )
-        headers[http_headers.BOS_RESTORE_TIER] = tier
+        if compat.convert_to_string(tier) not in ("Standard", "Expedited", "LowCost"):
+            raise ValueError('invalid tier:{} for restore_object.The valid value is \"Standard\" or \"Expedited\" or '\
+                    '\"LowCost\"'.format(tier) )
+        headers[http_headers.BOS_RESTORE_TIER] = compat.convert_to_bytes(tier)
         return self._send_request(
             http_methods.POST,
             bucket_name,
@@ -950,10 +1005,17 @@ class BosClient(BceBaseClient):
             body_parser=BosClient._parse_bos_object)
 
     @staticmethod
-    def _save_body_to_file(http_response, response, file_name, buf_size):
+    def _save_body_to_file(http_response, response, file_name, buf_size=16 * 1024, progress_callback=None):
         f = open(file_name, 'wb')
         try:
-            shutil.copyfileobj(http_response, f, buf_size)
+            # Added progress bar monitoring
+            if progress_callback:
+                file_size = int(response.metadata.content_length)
+                stream = utils.make_progress_adapter(http_response, progress_callback, file_size)
+            else:
+                stream = http_response
+
+            shutil.copyfileobj(stream, f, buf_size)
             http_response.close()
         finally:
             f.close()
@@ -981,7 +1043,8 @@ class BosClient(BceBaseClient):
         return s
 
     @required(bucket_name=(bytes, str), key=(bytes, str), file_name=(bytes, str))
-    def get_object_to_file(self, bucket_name, key, file_name, range=None, config=None):
+    def get_object_to_file(self, bucket_name, key, file_name, range=None, config=None, 
+                            progress_callback=None, traffic_limit=None):
         """
         Get Content of Object and Put Content to File
 
@@ -1000,18 +1063,28 @@ class BosClient(BceBaseClient):
             **HTTP Response**
         """
         key = compat.convert_to_bytes(key)
+        if len(key) == 0 or key.startswith(b"/"):
+            raise BceClientError("Key can not be empty or start with '/' .")
         file_name = compat.convert_to_bytes(file_name)
+        range_header = BosClient._get_range_header_dict(range)
+        if traffic_limit is not None:
+            if range_header is None:
+                range_header = {}
+            range_header[http_headers.BOS_TRAFFIC_LIMIT] = traffic_limit
+        
         return self._send_request(
             http_methods.GET,
             bucket_name,
             key,
-            headers=BosClient._get_range_header_dict(range),
+            headers=range_header,
             config=config,
             body_parser=lambda http_response, response: BosClient._save_body_to_file(
                 http_response,
                 response,
                 file_name,
-                self._get_config_parameter(config, 'recv_buf_size')))
+                self._get_config_parameter(config, 'recv_buf_size'),
+                progress_callback=progress_callback))
+
 
     @required(bucket_name=(bytes, str), key=(bytes, str))
     def get_object_meta_data(self, bucket_name, key, config=None):
@@ -1043,6 +1116,8 @@ class BosClient(BceBaseClient):
                      content_sha256=None,
                      storage_class=None,
                      user_headers=None,
+                     progress_callback=None,
+                     traffic_limit=None,
                      config=None):
         """
         Put an appendable object to BOS or add content to an appendable object
@@ -1067,7 +1142,8 @@ class BosClient(BceBaseClient):
             content_sha256=content_sha256,
             user_metadata=user_metadata,
             storage_class=storage_class,
-            user_headers=user_headers)
+            user_headers=user_headers,
+            traffic_limit=traffic_limit)
 
         if content_length > bos.MAX_APPEND_OBJECT_LENGTH:
             raise ValueError('Object length should be less than %d. '
@@ -1076,6 +1152,9 @@ class BosClient(BceBaseClient):
         params = {b'append': b''}
         if offset is not None:
             params[b'offset'] = offset
+        
+        if progress_callback:
+            data = utils.make_progress_adapter(data, progress_callback)
 
         return self._send_request(
             http_methods.POST,
@@ -1097,6 +1176,8 @@ class BosClient(BceBaseClient):
                                   content_sha256=None,
                                   storage_class=None,
                                   user_headers=None,
+                                  progress_callback=None,
+                                  traffic_limit=None,
                                   config=None):
         """
         Create an appendable object and put content of string to the object
@@ -1124,6 +1205,8 @@ class BosClient(BceBaseClient):
                                       content_sha256=content_sha256,
                                       storage_class=storage_class,
                                       user_headers=user_headers,
+                                      progress_callback=progress_callback,
+                                      traffic_limit=traffic_limit,
                                       config=config)
         finally:
             if fp is not None:
@@ -1145,6 +1228,8 @@ class BosClient(BceBaseClient):
                    encryption=None,
                    customer_key=None,
                    customer_key_md5=None,
+                   progress_callback=None,
+                   traffic_limit=None,
                    config=None):
         """
         Put object and put content of file to the object
@@ -1173,13 +1258,17 @@ class BosClient(BceBaseClient):
             content_sha256=content_sha256,
             user_metadata=user_metadata,
             storage_class=storage_class,
-            user_headers=user_headers)
+            user_headers=user_headers,
+            traffic_limit=traffic_limit)
 
         buf_size = self._get_config_parameter(config, 'recv_buf_size')
 
         if content_length > bos.MAX_PUT_OBJECT_LENGTH:
             raise ValueError('Object length should be less than %d. '
                              'Use multi-part upload instead.' % bos.MAX_PUT_OBJECT_LENGTH)
+        
+        if progress_callback:
+                data = utils.make_progress_adapter(data, progress_callback)
 
         return self._send_request(
             http_methods.PUT,
@@ -1200,6 +1289,8 @@ class BosClient(BceBaseClient):
                                encryption=None,
                                customer_key=None,
                                customer_key_md5=None,
+                               progress_callback=None,
+                               traffic_limit=None,
                                config=None):
         """
         Create object and put content of string to the object
@@ -1239,6 +1330,8 @@ class BosClient(BceBaseClient):
                                    encryption=encryption,
                                    customer_key=customer_key,
                                    customer_key_md5=customer_key_md5,
+                                   progress_callback = progress_callback,
+                                   traffic_limit=traffic_limit,
                                    config=config)
         finally:
             if fp is not None:
@@ -1256,7 +1349,10 @@ class BosClient(BceBaseClient):
                              encryption=None,
                              customer_key=None,
                              customer_key_md5=None,
-                             config=None):
+                             progress_callback=None,
+                             traffic_limit=None,
+                             config=None,
+                             ):
 
         """
         Put object and put content of file to the object
@@ -1299,6 +1395,8 @@ class BosClient(BceBaseClient):
                                    encryption=encryption,
                                    customer_key=customer_key,
                                    customer_key_md5=customer_key_md5,
+                                   progress_callback=progress_callback,
+                                   traffic_limit=traffic_limit,
                                    config=config)
         finally:
             fp.close()
@@ -1316,6 +1414,7 @@ class BosClient(BceBaseClient):
                     storage_class=None,
                     user_headers=None,
                     copy_object_user_headers=None,
+                    traffic_limit=None,
                     config=None):
         """
         Copy one object to another object
@@ -1340,7 +1439,8 @@ class BosClient(BceBaseClient):
             content_type=content_type,
             user_metadata=user_metadata,
             storage_class=storage_class,
-            user_headers=user_headers)
+            user_headers=user_headers,
+            traffic_limit=traffic_limit)
         headers[http_headers.BCE_COPY_SOURCE] = utils.normalize_string(
             b'/%s/%s' % (
                 compat.convert_to_bytes(source_bucket_name), 
@@ -1509,7 +1609,7 @@ class BosClient(BceBaseClient):
               part_fp=object)
     def upload_part(self, bucket_name, key, upload_id,
                     part_number, part_size, part_fp, part_md5=None,
-                    config=None):
+                    progress_callback=None, traffic_limit=None, config=None):
         """
         Upload a part.
 
@@ -1554,6 +1654,11 @@ class BosClient(BceBaseClient):
         if part_md5 is not None:
             headers[http_headers.CONTENT_MD5] = part_md5
 
+        if progress_callback:
+            part_fp = utils.make_progress_adapter(part_fp, progress_callback, part_size)
+        
+        if traffic_limit is not None:
+            headers[http_headers.BOS_TRAFFIC_LIMIT] = traffic_limit
         return self._send_request(
             http_methods.PUT,
             bucket_name,
@@ -1578,6 +1683,7 @@ class BosClient(BceBaseClient):
                          etag=None,
                          content_type=None,
                          user_metadata=None,
+                         traffic_limit=None,
                          config=None):
         """
         Copy part.
@@ -1604,7 +1710,8 @@ class BosClient(BceBaseClient):
         target_key = compat.convert_to_bytes(target_key)
         headers = self._prepare_object_headers(
                          content_type=content_type,
-                         user_metadata=user_metadata)
+                         user_metadata=user_metadata,
+                         traffic_limit=traffic_limit)
         headers[http_headers.BCE_COPY_SOURCE] = utils.normalize_string(
                          b"/%s/%s" % (compat.convert_to_bytes(source_bucket_name),
                          source_key), False)
@@ -1630,7 +1737,7 @@ class BosClient(BceBaseClient):
               offset=compat.integer_types)
     def upload_part_from_file(self, bucket_name, key, upload_id,
                               part_number, part_size, file_name, offset, part_md5=None,
-                              config=None):
+                              progress_callback=None, traffic_limit=None, config=None):
         """
 
         :param bucket_name:
@@ -1649,7 +1756,8 @@ class BosClient(BceBaseClient):
         try:
             f.seek(offset)
             return self.upload_part(bucket_name, key, upload_id, part_number, part_size, f,
-                                    part_md5=part_md5, config=config)
+                                    part_md5=part_md5, progress_callback=progress_callback,
+                                    traffic_limit=traffic_limit, config=config)
         finally:
             f.close()
 
@@ -1835,13 +1943,15 @@ class BosClient(BceBaseClient):
                 break
 
     def _upload_task(self, bucket_name, object_key, upload_id,
-        part_number, part_size, file_name, offset, part_list, uploadTaskHandle):
+        part_number, part_size, file_name, offset, part_list, uploadTaskHandle,
+        progress_callback=None, traffic_limit=None):
         if uploadTaskHandle.is_cancel():
             _logger.debug("upload task canceled with partNumber={}!".format(part_number))
             return
         try:
             response = self.upload_part_from_file(bucket_name, object_key, upload_id,
-                part_number, part_size, file_name, offset)
+                part_number, part_size, file_name, offset, progress_callback=progress_callback
+                , traffic_limit=traffic_limit)
             part_list.append({
                 "partNumber": part_number,
                 "eTag": response.metadata.etag
@@ -1858,6 +1968,8 @@ class BosClient(BceBaseClient):
             content_type=None,
             storage_class=None,
             user_headers=None,
+            progress_callback=None,
+            traffic_limit=None,
             config=None):
         """
         Multipart Upload file to bos
@@ -1868,9 +1980,9 @@ class BosClient(BceBaseClient):
         if chunk_size > 5 * 1024 or chunk_size <= 0:
            raise BceClientError("chunk size is valid, it should be more than 0 and not nore than 5120!")
         left_size = os.path.getsize(file_name)
-        # if file size more than 5TB, reject
-        if left_size > 5 * 1024 * 1024 * 1024 * 1024:
-           raise BceClientError("File size must not be more than 5TB!")
+        # if file size more than 48.8TB, reject
+        if left_size > 50000 * 1024 * 1024 * 1024:
+           raise BceClientError("File size must not be more than 48.8TB!")
         if thread_num is None or thread_num <= 1:
            thread_num = multiprocessing.cpu_count()
         part_size = chunk_size * 1024 * 1024
@@ -1895,7 +2007,7 @@ class BosClient(BceBaseClient):
             if left_size < part_size:
                 part_size = left_size
             temp_task= executor.submit(self._upload_task, bucket_name, key, upload_id, part_number, part_size,
-                file_name, offset, part_list, uploadTaskHandle)
+                file_name, offset, part_list, uploadTaskHandle, progress_callback, traffic_limit)
             all_tasks.append(temp_task)
             left_size -= part_size
             offset += part_size
@@ -2173,6 +2285,90 @@ class BosClient(BceBaseClient):
             )
         return select_response
 
+    def get_user_quota(self, config=None):
+        """
+        get user quota
+
+        :param config:
+        :return:
+        """
+        return self._send_request(
+            http_methods.GET, params={b'userQuota': b''})
+    
+    def put_user_quota(self, max_bucket_count, max_capacity_mega_bytes, config=None):
+        """
+        put user quota
+
+        :type max_bucket_count: int
+        :param max_bucket_count: max bucket count
+
+        :type max_capacity_mega_bytes: long
+        :param max_capacity_mega_bytes: max capacity mega bytes
+
+        :param config:
+        :return:
+        """
+        return self._send_request(
+            http_methods.PUT,
+            body=json.dumps({'maxBucketCount': max_bucket_count, 
+                            'maxCapacityMegaBytes': max_capacity_mega_bytes}),
+            params={b'userQuota': b''})
+
+    def delete_user_quota(self, config=None):
+        """
+        delete user quota
+
+        :param config:
+        :return:
+        """
+        return self._send_request(
+            http_methods.DELETE, params={b'userQuota': b''})        
+        
+    def get_notification(self, bucket_name, config=None):
+        """
+        get notification
+
+        :type bucket_name: string
+        :param bucket_name: bucket name
+
+        :param config:
+        :return:
+        """
+        return self._send_request(
+            http_methods.GET, bucket_name=bucket_name, params={b'notification': b''})
+    
+    def put_notification(self, bucket_name, notifications, config=None):
+        """
+        put user quota
+
+        :type bucket_name: string
+        :param bucket_name: bucket
+
+        :type notifications: list of dict
+        :param notifications: notifacation param
+
+        :param config:
+        :return:
+        """
+        return self._send_request(
+            http_methods.PUT, bucket_name=bucket_name,
+            body=json.dumps({'notifications': notifications}),
+            params={b'notification': b''})
+
+    def delete_notification(self, bucket_name, config=None):
+        """
+        delete notification
+
+        :type bucket_name: string
+        :param bucket_name: bucket name
+
+        :param config:
+        :return:
+        """
+        return self._send_request(
+            http_methods.DELETE, bucket_name=bucket_name, params={b'notification': b''})        
+        
+
     @staticmethod
     def _prepare_object_headers(
             content_length=None,
@@ -2185,7 +2381,8 @@ class BosClient(BceBaseClient):
             user_headers=None,
             encryption=None,
             customer_key=None,
-            customer_key_md5=None):
+            customer_key_md5=None,
+            traffic_limit=None):
         headers = {}
 
         if content_length is not None:
@@ -2241,6 +2438,9 @@ class BosClient(BceBaseClient):
                 headers = BosClient._get_user_header(headers, user_headers, False)
             except Exception as e:
                 raise e
+        
+        if traffic_limit is not None:
+            headers[http_headers.BOS_TRAFFIC_LIMIT] = traffic_limit
 
         return headers
 
@@ -2281,17 +2481,29 @@ class BosClient(BceBaseClient):
         host = config.endpoint
         if use_backup_endpoint:
             host = config.backup_endpoint
-        if config.cname_enabled or utils.is_cname_like_host(host):
+        if config.cname_enabled or utils.is_cname_like_host(host) or utils.is_custom_host(host, bucket_name):
             return utils.append_uri(bos.URL_PREFIX, key)
         return utils.append_uri(bos.URL_PREFIX, bucket_name, key)
 
-    def _merge_config(self, config):
-        if config is None:
-            return self.config
-        else:
-            new_config = copy.copy(self.config)
+    def _merge_config(self, config, bucket_name):
+        # if config is None:
+        #     return self.config
+        # else:
+        #     new_config = copy.copy(self.config)
+        #     new_config.merge_non_none_values(config)
+        #     return new_config
+        
+        new_config = copy.copy(self.config)
+        if config is not None:
             new_config.merge_non_none_values(config)
-            return new_config
+        if bucket_name is not None and not utils.is_cname_like_host(self.config.endpoint):
+            user_endpoint = self.config.endpoint
+            user_endpoint_split_size = len(compat.convert_to_bytes(user_endpoint).split(b'.'))
+            if user_endpoint.endswith(DEFAULT_BOS_DOMAIN_SUFFIX) and user_endpoint_split_size == 3:
+                new_config.endpoint = compat.convert_to_bytes(bucket_name)+b'.'+\
+                compat.convert_to_bytes(user_endpoint)
+
+        return new_config
 
     @staticmethod
     def _need_retry_backup_endpoint(error):
@@ -2314,7 +2526,8 @@ class BosClient(BceBaseClient):
             body=None, headers=None, params=None,
             config=None,
             body_parser=None):
-        config = self._merge_config(config)
+        config = self._merge_config(config, bucket_name)
+
         path = BosClient._get_path(config, bucket_name, key)
         if body_parser is None:
             body_parser = handler.parse_json
